(***********************************************************************)
(* Add-ons                                                             *)
(* Bruno Blanchet, CNRS / Laboratoire d'Informatique                   *)
(*                 de l'cole Normale Suprieure                       *)
(* David Monniaux, Laboratoire d'Informatique de                       *)
(*                 l'cole Normale Suprieure                          *)
(* Antoine Mine',  Laboratoire d'Informatique de                       *)
(*                 l'cole Normale Suprieure                          *)
(* Xavier Rival,   Laboratoire d'Informatique de                       *)
(*                 l'cole Normale Suprieure                          *)
(***********************************************************************)

(* $Id: map2.mli,v 1.1 2008-04-30 15:25:18 monniaux Exp $ *)

(** Association tables over ordered types.

   This module implements applicative association tables, also known as
   finite maps or dictionaries, given a total ordering function
   over the keys.
   All operations over maps are purely applicative (no side-effects).
   The implementation uses balanced binary trees, and therefore searching
   and insertion take time logarithmic in the size of the map. 
*)

module type OrderedType = 
  sig
    type t
      (** The type of the map keys. *)
    val compare : t -> t -> int
      (** A total ordering function over the keys.
          This is a two-argument function [f] such that
          [f e1 e2] is zero if the keys [e1] and [e2] are equal,
          [f e1 e2] is strictly negative if [e1] is smaller than [e2],
          and [f e1 e2] is strictly positive if [e1] is greater than [e2].
          Example: a suitable ordering function is
          the generic structural comparison function {!Pervasives.compare}. *)
  end
(** Input signature of the functor {!Map.Make}. *)

type ('key,'data) node = 
    Empty 
  | Node of ('key,'data) node * 'key * 'data * ('key,'data) node * int
(* non-abstract type to ease debugging [AM] *)

module Make (Ord : OrderedType) : Maps.MAP_TYPE with type key = Ord.t
and type 'a t = (Ord.t,'a) node
(** Functor building an implementation of the map structure
   given a totally ordered type. *)

